'******************************************************************************
' File: RapidDBG.bas - part of RapidDBG, a RapidQ source-line debugger.
' Copyright (C) 2008-2013 Don Homans
'
'    RapidDBG is free software: you can redistribute it and/or modify
'    it under the terms of the GNU General Public License as published by
'    the Free Software Foundation, either version 3 of the License, or
'    (at your option) any later version.
'
'    RapidDBG is distributed in the hope that it will be useful,
'    but WITHOUT ANY WARRANTY; without even the implied warranty of
'    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'    GNU General Public License for more details.
'
'    You should have received a copy of the GNU General Public License
'    along with RapidDBG.  If not, see <http://www.gnu.org/licenses/>.
'
'******************************************************************************
'  PROJECT: RapidDBG List of input files
'      RapidDBG.bas: A small GUI wrapper that controls the generation of a
'               temporary debug+src executable and starts the debugging process.
'               It includes a debug source code color editor (-E switch).
'      RapidDBG.inc : Bug fix for QColorDialog in Rapidq2.inc
'      RapidDBG.dat : The debugger main GUI and logic.
'      RapidDBG.bmp : Default toolbar bitmap.
'      RapidDBG.ico : Default project icon.
'
'      RapidDBG.fbb : A FreeBasic dll that parses the user src file, adding debug
'                    information, into a temporary debug+src RapidQ source file.
'
'      CodeStub.txt : Demo code for the source code color editor.
'      Whats New.txt: Some helpful(?) information on updates etc.
'
'           RapidDBG List of output files
'      RapidDBG.exe : Main entry to the debugger
'      RapidDBG.dll : Parser library
'
'      AUTHOR: d_homans@yahoo.com.au
'     VERSION: v1.53 (defined in RapidDBG.fbb)
'        DATE: 25 May 2013
' CONTRIBUTOR: Jacques Philippe
'------------------------------------------------------------------------------------------

$APPTYPE GUI
$TYPECHECK ON
$OPTION ICON "RapidDBG.ico"                                         ' Using the RapidQ icon

$INCLUDE <Rapidq2.inc>
$INCLUDE "RapidDBG.inc"

$DEFINE TEST_ONLY   1                                               ' -t/-T  Syntax check - compile file, delete exe and display results
$DEFINE COMP_ONLY   2                                               ' -c/-C  Compile file to exe - display (error) result
$DEFINE RUN_USER    3                                               ' -r/-R  Compile and run file
$DEFINE RUN_DEBUG   4                                               ' -d/-D  Add debug code, compile then run the debugger
$DEFINE EDIT_COLORS 5                                               ' -e/-E  Edit color scheme of debugger display
$DEFINE CLEANUP_DBG 6                                               ' -x/-X  "Internal" cmd - deletes debugger exe file

$DEFINE DBG_MOUSEWHEEL  &H20A

CONST CRLF=CHR$(13)+CHR$(10)
CONST QU=CHR$(34)

$RESOURCE DebugData AS "RapidDBG.dat"                               ' |
$RESOURCE ProgIcon AS "RapidDBG.ico"                                ' | Order here important for code below
$RESOURCE ButnPics AS "RapidDBG.bmp"                                ' |
$RESOURCE Template AS "CodeStub.txt"                                ' |

DECLARE FUNCTION WaitOnCompile LIB "kernel32" ALIAS "WaitForSingleObject" (hHandle AS long, _
                                                                        dwMilliseconds AS long) AS long
DECLARE FUNCTION WindowFromPoint LIB "user32" ALIAS "WindowFromPoint" (X AS long, Y AS long) AS long

DECLARE FUNCTION fbParserVersion LIB "RapidDBG" ALIAS "fbParserVersion@4" (lpRetString AS long) AS integer
DECLARE FUNCTION fbParseFile LIB "RapidDBG" ALIAS "fbParseFile@20" (lpSrcFile AS long, lpDataFile AS long, _
                                    lpInitData AS long, lpRetString AS long, buflen AS integer) AS integer


DECLARE SUB DbgNoClose(Action AS integer)
DECLARE FUNCTION TRIM$(srcLine AS string) AS string
' Initialise fns
DECLARE SUB Initialise                                              ' Some init stuff
DECLARE SUB GetEnvironment                                          ' User input -> Paths to rc.exe etc....
DECLARE SUB LoadEnvironment                                         ' ....or Load pre-set paths
' GP subs/fns
DECLARE SUB KillDebugEXE
DECLARE SUB DbgError(ErrNum AS integer)
DECLARE SUB GotResponse(Sender AS QBUTTON)
DECLARE SUB WaitResponse(msg AS string, msgType AS integer)
' In-line steps to debug a file
DECLARE SUB TestUserSource                                          ' Ensure user source file compiles ok!
DECLARE SUB CreateDbgTarget
DECLARE SUB CreateDbgExe                                            ' Compile to final exe with debug code
DECLARE SUB RunDbgExe                                               ' Run final exe and terminate

' Editor for displayed source code colour scheme
DECLARE SUB InitEdit
DECLARE SUB LoadColourScheme
DECLARE SUB IdentifyEachLine
DECLARE SUB ColourThisLine(Col AS integer,Row AS integer,State AS integer,Rect AS QRECT)
DECLARE SUB EditColour(Sender AS QPANEL)
DECLARE SUB ShowHelp
DECLARE SUB EditAbort
DECLARE SUB EditExit

'--------------------------Global data-------------------------------------------------------------------
DEFSTR VERSION                                                      ' Obtained from RapidDBG.dll

DEFLNG OldWndproc=0                                                 ' Provide mouse scroll for color editor
DEFINT CommandParm                                                  ' Command-line parameter
DEFSTR rcPath                                                       ' Our environ variables....
DEFSTR incPath                                                      ' ... note:  no "\" terminator
DEFSTR libPath
DEFSTR iniFile                                                      ' Environment variables file
DEFSTR srcFile                                                      ' Test source file
DEFSTR dbgFile                                                      ' Test src plus debug info
DEFSTR PreSetBrkPoints                                              ' (Command line option) preset breakpoints

DIM SrcStream AS QMEMORYSTREAM                                      ' Source file passed to FB dll
DIM DatStream AS QMEMORYSTREAM                                      ' Data passed to FB dll
DIM tmpStream AS QMEMORYSTREAM

' v1.43 Colour editor variables
DIM Label(1 TO 10) AS QLABEL
DIM FGPanel(1 TO 10) AS QPANEL
DIM BGPanel(1 TO 10) AS QPANEL
DEFINT FGColor(1 TO 10), BGColor(1 TO 10)                           ' v1.51 Arrays ilo separate vars
DIM DisplayList AS QSTRINGLIST
DIM ColorDlg AS QColorDlg

'--------------------------------Forms------------------------------------------------------------
CREATE DbgForm AS QFORM                                             ' Main form
    Width=500
    Height=200
    DelBorderIcons(2)                                               ' Disable maximize button
    BorderStyle=1                                                   ' Fixed size form
    Font.Size=9
    Font.AddStyles fsBold
    Center
    Caption="RapidDBG "+VERSION
    OnClose=DbgNoClose                                              ' Disable "X" box of form
    CREATE RichEd AS QRichEdit
        Align=5
        HideScrollBars=False
        ScrollBars=ssVertical
        PlainText=True
        ReadOnly=True
        Alignment=0
        HideSelection=True
    END CREATE
    CREATE DirTree AS QDirTree
        Align=5
        Visible=False
        InitialDir=LEFT$(Application.Path,3)
    END CREATE
    CREATE OKButn AS QButton
        Left=(DbgForm.ClientWidth-OKButn.Width)\2
        Top=DbgForm.ClientHeight-OKButn.Height
        Caption="Set"
        Tag=0
        Visible=False
        OnClick=GotResponse
    END CREATE
END CREATE

' Source code colour editor form
CREATE EditForm AS QFORM
    Width=720
    Height=480
    BorderStyle=bsSingle
    Caption="Debugger Color Editor"
    DelBorderIcons(biMinimize,biMaximize)                          ' Disable min/max buttons
    Center
    CREATE StrGrid AS QSTRINGGRID
        Width=500
        Height=EditForm.ClientHeight*7\8                            ' v1.52 addition
        Left=200
        Top=10
        FixedRows=1
        FixedCols=0
        ColCount=3
        RowCount=100
        DefaultRowHeight=16
        Font.Name="Courier"
        ColWidths(0)=20
        ColWidths(1)=35
        ColWidths(2)=425
        ScrollBars=ssVertical
        ColumnStyle(0)=gcsNone
        ColumnStyle(1)=gcsNone
        ColumnStyle(2)=gcsNone
        Color=&hffffff                                              ' White background
        GridLineWidth=0
        Cell(0,0)="BP"
        Cell(1,0)="Line"
        Cell(2,0)="File: Source code example"
        OnDrawCell=ColourThisLine                                   ' Updates colour of text on screen
    END CREATE
    CREATE Lbl1 AS QLABEL
        Left=140
        Top=10
        Caption="Select"
    END CREATE
    CREATE Btn1 AS QBUTTON
        Width=100
        Left=EditForm.Width\4-Btn1.Width\2
        Top=EditForm.ClientHeight*15\16                             ' v1.52
        Caption="SAVE && EXIT"
        OnClick=EditExit
    END CREATE
    CREATE Btn2 AS QBUTTON
        Width=50
        Left=(EditForm.Width-Btn2.Width)\2
        Top=Btn1.Top
        Caption="HELP"
        OnClick=ShowHelp
    END CREATE
    CREATE Btn3 AS QBUTTON
        Width=100
        Left=EditForm.Width*3\4-Btn3.Width\2
        Top=Btn1.Top
        Caption="CANCEL"
        OnClick=EditAbort
    END CREATE
END CREATE

Initialise

SELECT CASE CommandParm
    CASE EDIT_COLORS
        InitEdit                                                    ' Customize debugger screen
        EditForm.ShowModal
    CASE CLEANUP_DBG
        KillDebugEXE                                                ' Cleanup debugger exe file
    CASE ELSE
        TestUserSource                                              ' -T -C -R exit on completion
        CreateDbgTarget                                             ' -D option Create debug+src bas file
        CreateDbgExe                                                ' Compile debug+src in the target folder
        RunDbgExe                                                   ' Start debugging
        DbgError(0)                                                 ' Finished
END SELECT
END

SUB DbgNoClose(Action)
	Action=False                                                    ' No closing of form
END SUB

FUNCTION TRIM$(srcLine)
    Result=RTRIM$(LTRIM$(srcLine))
END FUNCTION

'-------------------------- Calling the FreeBasic dll-----------------------------------
' The Dll creates fn/var map in memory and then builds a debug+src file ready to compile
' Params [5]: src file (memorystream), data file (memorystream), initData address of string,
' address of return buffer, size of buffer
' Returns: Fail 0 (reason in buffer?) OR Pass 1
'---------------------------------------------------------------------------------------
SUB CreateDbgTarget
DEFINT i, retVal
DEFSTR buf, tstr
DIM tFile AS QFileStream

    RichEd.AddStrings "Parsing "+srcFile
    buf=SPACE$(256)                                                 ' Returned string from dll call
    SrcStream.Close                                                 ' Initialise it
    tFile.Open (srcFile, fmOpenRead)
    SrcStream.CopyFrom(tFile,0)                                     ' Get the source file into memory
    SrcStream.WriteNum(0,4)                                         ' Append some nulls BUGFIX: v1.35
    tFile.Close
    DoEvents
    SrcStream.Position=0                                            ' Not necessary but..
    DatStream.Close                                                 ' Initialise it
    DatStream.ExtractRes(Resource(0))                               ' Get included dat file in memory
    DatStream.WriteNum(0,4)                                         ' Append some nulls BUGFIX: v1.35
    DatStream.Position=0

    tstr=srcFile+"~"                                                ' Full path+src file
    tstr=tstr+Application.Path+"~"                                  ' Path to this program
    tstr=tstr+PreSetBrkPoints+"~"                                   ' Debugger preset breakpoints (if any)
    tstr=tstr+STR$(DbgForm.Left)+"~"                                ' DLL Progress bar Left
    tstr=tstr+STR$(DbgForm.Top+DbgForm.Height-50)+"~"               ' DLL Progress bar Top

    retVal=fbParseFile(SrcStream.Pointer, DatStream.Pointer, VARPTR(tstr), VARPTR(buf), LEN(buf))

    SrcStream.Close: DatStream.Close: tmpStream.Close
    IF retVal=0 THEN
        tstr="ERROR: CreateDbgTarget()"+CRLF+"Reason:"+CRLF+RTRIM$(buf)+CRLF+"Terminating!"
        MessageBox(tstr,"RapidDBG "+VERSION,&H10)
        Application.Terminate
    END IF
'    MessageBox(buf,"Information",&H40)
    buf=" ": tstr=" "
END SUB

'------------------------------Test/RUN User source--------------------------------------------
' Entry point TestUserSource()

' Debugging - NOT interested in code with line numbers
FUNCTION HasLineNumbers() AS integer
DEFINT i
DEFSTR tstr
DIM tFile AS QFileStream

    tmpStream.Close                                                 ' Re-initialise it
    tFile.Open(srcFile,fmOpenRead)                                  ' Open/Read source file
    tmpStream.CopyFrom(tFile,0)                                     ' Copy the entire stream
    tFile.Close
    DoEvents
    tmpStream.Position=0
    FOR i=1 TO 32                                                   ' Check 1st 32 lines of source
        tstr=tmpStream.ReadLine
        tstr=REPLACESUBSTR$(tstr,CHR$(9)," ")                       ' Replace TABs
        IF LEFT$(UCASE$(tstr),4)<>"REM " AND LEFT$(tstr,1)<>"'" THEN
            tstr=FIELD$(tstr," ",1)
            IF VAL(tstr)>0 THEN                                     ' Line number?
                Result=True
                EXIT FUNCTION
            END IF
        END IF
    NEXT i
    Result=False
END FUNCTION

SUB RunUserProgram
DEFSTR tgtFile

    tgtFile=LEFT$(srcFile,RINSTR(srcFile,"."))+"EXE"                ' The EXE tgtFile
    IF FileExists(tgtFile)=0 THEN
        MessageBox(tgtFile+" - File does not exists!","RapidDBG "+VERSION,&H10)
    ELSE
        RUN tgtFile                                                 ' "Run" the program and return to quit
    END IF
    Application.Terminate
END SUB

FUNCTION CompileFile(FileStr AS string) AS integer
DEFINT size
DEFLNG PID
DEFSTR tgtFile, RQCompile
DIM File AS QFILESTREAM

    tgtFile=LEFT$(FileStr,RINSTR(FileStr,"."))+"EXE"
    RQCompile=QU+rcPath+"\RC.EXE"+QU+" -r "+QU+"-I"+incPath+QU+" "
    RQCompile=RQCompile+QU+"-L"+libPath+QU+" "+QU+FileStr+QU+" "+QU+tgtFile+QU

    Result=False                                                    ' Default fail
    PID=SHELL(RQCompile,0)
    WaitOnCompile(PID,60000)                                        ' 60 secs?
    IF FileExists(tgtFile) THEN
        File.Open(tgtFile,fmOpenRead)
        size=File.Size
        File.Close
        DoEvents
        IF size>0 THEN Result=True
    END IF
END FUNCTION

' Check syntax of source file: Take appropriate action
' Action dependent on command line switch
SUB TestUserSource
DEFINT i
DEFSTR tstr, RQDumpStr
DIM tFile AS QFileStream

    RichEd.AddStrings "Checking syntax of "+srcFile
    IF CommandParm=RUN_DEBUG THEN
        IF HasLineNumbers() THEN DbgError(1)                        ' EXIT immediately
    END IF
    
    RQDumpStr=LEFT$(srcFile,RINSTR(srcFile,"\"))+"DUMP.$$$"
    IF CompileFile(srcFile) THEN                                    ' Compile ok?
        KILL RQDumpStr                                              ' Delete - not needed
        RichEd.AddStrings "Success!"
        DoEvents
        tstr=LEFT$(srcFile,RINSTR(srcFile,"."))+"EXE"               ' EXE file created
        SELECT CASE CommandParm
            CASE TEST_ONLY                                          ' Syntax check user source code
                KILL tstr                                           ' Delete exe - not wanted
                DoEvents
                MessageBox("Your code compiles ok!","RapidDBG "+VERSION,&H40)
                Application.Terminate
            CASE COMP_ONLY                                          ' Compile to exe
                MessageBox("Your code compiled ok!","RapidDBG "+VERSION,&H40)
                Application.Terminate
            CASE RUN_USER                                           ' Compile and run program
                RichEd.AddStrings "Running program..."
                DoEvents
                RunUserProgram                                      ' Run exe and terminate
            CASE ELSE                                               ' Syntax check before debugging
                KILL tstr                                           ' Delete exe - not wanted
                DoEvents
        END SELECT
    ELSE
        tFile.Open(RQDumpStr,fmOpenRead)
        tstr=tFile.ReadLine                                         ' Junk first line
        tstr="Rapid-Q Compiler Beta by William Yu"+CRLF
        tstr=tstr+"Copyright 1999-2000"+CRLF+CRLF
        tstr=tstr+"______________________________________________"+CRLF
        FOR i=1 TO tFile.LineCount-1                                ' Lines terminated with LF char...
            tstr=tstr+tFile.ReadLine+CHR$(13)                       ' ...so MUST add termination...
        NEXT i                                                      ' ... to display in our RichEdit
        tstr=tstr+CRLF
        tstr=tstr+"____________________.o0o.______________________"+CRLF

        tFile.Close                                                 ' Leave the "DUMP" file behind!
        DoEvents
        MessageBox(tstr,"RapidDBG "+VERSION,&H10)
        Application.Terminate
    END IF
END SUB

'----------------------------------Compile Debug+source--------------------------------------
' Place a copy of debugger icon & toolbar bmp into target folder
SUB CreateDebugImages
DIM tFile AS QFileStream
DEFSTR tstr

    tstr=LEFT$(dbgFile,RINSTR(dbgFile,"\"))                         ' Debug destination folder
    tFile.Open(tstr+"RapidDBG.ico",fmCreate)
    tFile.ExtractRes(RESOURCE(1))                                   ' Icon image
    tFile.Close
    DoEvents
    tFile.Open(tstr+"RapidDBG.bmp",fmCreate)
    tFile.ExtractRes(RESOURCE(2))                                   ' Toolbar image
    tFile.Close
    DoEvents
END SUB

' On failure, remove these from target folder
SUB DeleteDebugImages
DEFSTR tstr

    tstr=LEFT$(dbgFile,RINSTR(dbgFile,"\"))                         ' Debug destination folder
    KILL tstr+"RapidDBG.ico"
    KILL tstr+"RapidDBG.bmp"
    DoEvents
END SUB

SUB CreateDbgExe
DEFINT i
DEFSTR tstr, RQDumpStr
DIM tFile AS QFileStream

    RichEd.AddStrings "Generating debug executable"
    DoEvents
    CreateDebugImages                                               ' NEED these to compile
    RQDumpStr=LEFT$(dbgFile,RINSTR(dbgFile,"\"))+"DUMP.$$$"
    IF CompileFile(dbgFile)=False THEN
        tFile.Open(RQDumpStr,fmOpenRead)

        tstr="_____________### RapidDBG FAILURE ###_____________"+CRLF+CRLF
        FOR i=1 TO tFile.LineCount
            tstr=tstr+tFile.ReadLine+CHR$(13)                       ' MUST add <CR> termination char
        NEXT i
        tstr=tstr+CRLF+"Please advise the author of this problem."+CRLF
        tstr=tstr+"Please include an extract of the code"+CRLF
        tstr=tstr+"around the error line in file:"+CRLF+dbgFile+CRLF
        tstr=tstr+"_______________________.o0o._______________________"+CRLF

        tFile.Close
        RichEd.AddStrings "### RapidDBG FAILURE ###"
        DoEvents
        MessageBox(tstr,"RapidDBG "+VERSION,&H10)
        DeleteDebugImages                                           ' Delete these
        Application.Terminate                                       ' Exit - Leaving DUMP.$$$ for reference
    END IF
    KILL RQDumpStr                                                  ' Dont need DUMP.$$$
    DoEvents
END SUB

'----------------------------------Run Compiled Debug+source--------------------------------------
' Start debugging target source file
SUB RunDbgExe
DEFSTR tgtFile

    RichEd.AddStrings "Running debugger...",""
    DoEvents
    Sleep.MS 250
    tgtFile=LEFT$(dbgFile,RINSTR(dbgFile,"."))+"EXE"
    IF FileExists(tgtFile) THEN
        RUN tgtFile
        Application.Terminate                                       ' We are done for now
    ELSE
        DbgError(8)
    END IF
END SUB

'----------------------------------General functions--------------------------------------
SUB DbgError(errnum)
DEFSTR tstr, msg="ERR #"+STR$(errnum)+CRLF
DEFINT msgIcon=&h10                                                 ' Red cross icon

    SELECT CASE errnum
        CASE 0: tstr="RapidDBG will now terminate": msg="": msgIcon=&h40    ' "i" information icon
        CASE 1: tstr="Line numbers are not supported in this version!": msgIcon=&h40
        CASE 2: tstr="RapidDBG ini file corrupt or missing!"
        CASE 3:
            tstr="USAGE:"+CRLF+"RapidDBG [<option> [<source file> [-b <Breakpoints>]]]"+CRLF+CRLF
            tstr=tstr+"<source file> must include full path, and be of type bas or rqb"+CRLF+CRLF
            tstr=tstr+"option:"+CRLF+"        None    Select a source file to debug"+CRLF
            tstr=tstr+"        -e/-E    Edit color scheme of the debugger"+CRLF+CRLF
            tstr=tstr+"        -t/-T    Check syntax of <source file>"+CRLF
            tstr=tstr+"        -c/-C    Compile <source file> to exe"+CRLF
            tstr=tstr+"        -r/-R    Compile then run <source file>"+CRLF
            tstr=tstr+"        -d/-D    Compile <source file> with debug code."+CRLF
            tstr=tstr+"                     Run the debugger."+CRLF+CRLF
            tstr=tstr+"        -b/-B    Optional preset breakpoints used with -d/-D"+CRLF
            tstr=tstr+"                    String of comma separated line numbers"+CRLF+CRLF
            msgIcon=&h40
        CASE 4: tstr="Your source file is invalid or non-existent": msgIcon=&h20
        CASE 5: tstr="Expecting BAS or RQB source file ONLY!": msgIcon=&h20             ' "?" icon
        CASE 6: tstr="FreeQ ini file corrupt!"
        CASE 7: tstr="Problem loading RapidDBG parser dll"
        CASE 8: tstr="RunDbgExe() - Internal Error!"
        CASE 9: tstr="Failed to get dll version info!"+CRLF+"RapidDBG terminating"
        CASE ELSE
            tstr="Unidentified Error!!!!"
    END SELECT
    
    msg=msg+tstr
    MessageBox(msg,"RapidDBG"+VERSION, msgIcon)
    Application.Terminate                                           ' Tidy up and exit
END SUB

' The -x/X parameter is an "internal" command:
' Debug session, on termination, calls RapidqDBG.exe with the -X switch
' and it's OWN exe filename. RapidDBG simply deletes the debug exe file.
SUB KillDebugEXE
DEFSTR tgtFile

    RichEd.Clear
    RichEd.AddStrings "RapidDBG "+VERSION+" - Terminating"
    DoEvents
    Sleep.MS 1000                                                   ' Allow time for tgtFile to terminate
    tgtFile=UCASE$(TRIM$(COMMAND$(2)))
    KILL tgtFile                                                    ' Fails until TgtFile terminated
    WHILE FileExists(tgtFile)
        DoEvents
        Sleep.MS 100
        KILL tgtFile                                                ' Fails until TgtFile terminated
    WEND
    Application.Terminate
END SUB

SUB GotResponse(Sender)
    Sender.Tag=1                                                    ' OKBtn clicked
END SUB

SUB WaitResponse(msg, msgType)
    OKButn.Tag=0                                                    ' Used for looping
    OKButn.Visible=True
    MessageBox(msg,"RapidDBG "+VERSION,msgType)
    WHILE OKButn.Tag=0
        DoEvents
    WEND
    OKButn.Tag=0                                                    ' Used for looping
    OKButn.Visible=False
END SUB

SUB CreateNewIniFile
DIM iniList AS QStringList

    iniList.Clear
    iniList.AddItems "[Environment]", "RapidqPath="+rcPath
    iniList.AddItems "RapidqInc="+incPath, "RapidqLib="+libPath
    iniList.AddItems "[Screen]"
    iniList.AddItems "FG_DEFAULT=0                  ' Default", "BG_DEFAULT=&Hffffff"
    iniList.AddItems "FG_CURLINE=&H007f00           ' Current Line", "BG_CURLINE=&Hbfffd7"
    iniList.AddItems "FG_BRKPOINT=&H0000ff          ' Breakpoint", "BG_BRKPOINT=&Hbfbfff"
    iniList.AddItems "FG_COMMENT=&Hafafaf           ' Comment", "BG_COMMENT=&Hf7f7f7"
    iniList.AddItems "FG_RQDEFS=0                   ' RQ DEf, CONST etc", "BG_RQDEFS=&Hffffff"
    iniList.AddItems "FG_RQCREATE=0                 ' RQ object", "BG_RQCREATE=&Hffffff"
    iniList.AddItems "FG_RQTYPES=0                  ' Structure (Type)", "BG_RQTYPES=&Hffffff"
    iniList.AddItems "FG_RQVARS=0                   ' Variable", "BG_RQVARS=&Hffffff"
    iniList.AddItems "FG_SUBFNS=0                   ' Start + End Sub/Func", "BG_SUBFNS=&Hffffff"
    iniList.AddItems "FG_SUBFNCODE=0                ' Code in Sub & Func", "BG_SUBFNCODE=&Hffffff"
    iniList.SaveToFile(Application.Path+"\RapidDBG.ini")
    iniList.Clear
END SUB

SUB GetEnvironment
    iniFile=Application.Path+"\RapidDBG.ini"
    DbgForm.Caption="Pre-set your RapidQ environment"
    DbgForm.Show
    DirTree.Visible=True

    WaitResponse("Set path to the RapidQ compiler",&h20)
    rcPath=LEFT$(DirTree.Directory,LEN(DirTree.Directory)-1)        ' Remove trailing "\"
    WaitResponse("Set path to RapidQ INCLUDE files",&h20)
    incPath=LEFT$(DirTree.Directory,LEN(DirTree.Directory)-1)
    WaitResponse("Set path to RapidQ LIBRARY files",&h20)
    libPath=LEFT$(DirTree.Directory,LEN(DirTree.Directory)-1)

    DirTree.Visible=False
    CreateNewIniFile
END SUB

SUB LoadDbgEnvironment
DIM iniList AS QStringList
DEFSTR tstr

    iniFile=Application.Path+"\RapidDBG.ini"
    iniList.Clear
    iniList.LoadFromFile(iniFile)
    IF iniList.ItemCount>3 THEN                                    ' New format
        IF UCASE$(iniList.Item(0))<>"[ENVIRONMENT]" THEN DbgError(2)
        rcPath=FIELD$(iniList.Item(1),"=",2)
        incPath=FIELD$(iniList.Item(2),"=",2)
        libPath=FIELD$(iniList.Item(3),"=",2)
    ELSE                                                            ' Old format?
        rcPath=iniList.Item(0)
        incPath=iniList.Item(1)
        libPath=iniList.Item(2)
        CreateNewIniFile                                            ' Re-write new format
    END IF
END SUB

SUB LoadIdeEnvironment
DEFINT i, cnt=0
DEFSTR tstr, token, drv
DIM iniList AS QStringList

    tstr=Application.Path
    drv=FIELD$(tstr,"\",1)
    iniFile=LEFT$(tstr,RINSTR(tstr,"\"))+"Freeq.ini"

    rcPath="": incPath="": libPath=""
    iniList.Clear
    iniList.LoadFromFile(iniFile)
    FOR i=0 TO iniList.ItemCount-1
        tstr=UCASE$(iniList.Item(i))
        tstr=REPLACESUBSTR$(tstr,CHR$(9)," ")                       ' Replace TABs
        tstr=REPLACESUBSTR$(tstr,"/","\")                           ' Just in case
        token=TRIM$(FIELD$(tstr,"=",1))
        SELECT CASE token
            CASE "COMPILERPATH": rcPath=TRIM$(FIELD$(tstr,"=",2)): INC cnt
            CASE "COMPILERLIBPATH": libPath=TRIM$(FIELD$(tstr,"=",2)): INC cnt
            CASE "COMPILERINCPATH": incPath=TRIM$(FIELD$(tstr,"=",2)): INC cnt
        END SELECT
        IF cnt=3 THEN EXIT FOR
    NEXT i
    IF (rcPath="") OR (libPath="") OR (incPath="") THEN DbgError(6)

    IF INSTR(rcPath,":")=0 THEN
        IF LEFT$(rcPath,1)="\" THEN rcPath=drv+rcPath ELSE rcPath=drv+"\"+rcPath
    END IF
    IF RIGHT$(rcPath,3)="EXE" THEN rcPath=LEFT$(rcPath,RINSTR(rcPath,"\")-1)

    IF INSTR(libPath,":")=0 THEN
        IF LEFT$(libPath,1)="\" THEN libPath=drv+libPath ELSE libPath=drv+"\"+libPath
    END IF
    IF RIGHT$(libPath,1)="\" THEN libPath=LEFT$(libPath,LEN(libPath)-1)

    IF INSTR(incPath,":")=0 THEN
        IF LEFT$(incPath,1)="\" THEN incPath=drv+incPath ELSE incPath=drv+"\"+incPath
    END IF
    IF RIGHT$(incPath,1)="\" THEN incPath=LEFT$(incPath,LEN(incPath)-1)

    CreateNewIniFile                                                ' Write new format
END SUB

SUB CheckEnvironment
DEFLNG hDll=0
DEFSTR tstr

    tstr=Application.Path
    tstr=LEFT$(tstr,RINSTR(tstr,"\"))+"Freeq.ini"                   ' Up one folder
    IF FileExists(Application.Path+"\RapidDBG.ini") THEN
        LoadDbgEnvironment											' RapidDBG ini file exists - load it
    ELSEIF FileExists(tstr) THEN
        LoadIdeEnvironment                                          ' Load from the IDE ini file
    ELSE
        GetEnvironment                                              ' else create an ini file!
    END IF

    IF FileExists(Application.Path+"\RapidDBG.dll")>0 THEN
        hDll=LoadLibrary(Application.Path+"\RapidDBG.dll")         ' Current path?
    ELSEIF FileExists(LibPath+"\RapidDBG.dll")>0 THEN
        hDll=LoadLibrary(LibPath+"\RapidDBG.dll")                  ' RQ lib path?
    END IF
    IF hDll=0 THEN DbgError(7)
END SUB

' Debug optional param:  -b/B brkpt1,brkpt2,....
FUNCTION CheckForBreakpoints() AS string
DEFSTR tstr

    tstr=""
    IF (CommandCount=4) AND (UCASE$(TRIM$(COMMAND$(3)))="-B") THEN
        tstr=TRIM$(COMMAND$(4))
    END IF
    Result=tstr
END FUNCTION

SUB GetSourceFile
DIM fDialog as QOpenDialog
DEFSTR fName, tgtFolder

    srcFile=""
    dbgFile=""
    fDialog.Caption="Select a Source File"
    fDialog.FileName=""
    fDialog.InitialDir=Application.Path
    fDialog.Filter="Source Files (*.bas *.rqb)|*.bas;*.rqb"
    IF fDialog.Execute THEN
        srcFile=fDialog.FileName
        fName=RIGHT$(srcFile,LEN(srcFile)-RINSTR(srcFile,"\"))
        tgtFolder=LEFT$(srcFile,RINSTR(srcFile,"\"))
        dbgFile=tgtFolder+"~"+fName
        CommandParm=RUN_DEBUG
    ELSE
        RichEd.AddStrings "No source file selected!"
        DbgError(0)                                                 ' No file selected, just quit!
    END IF
END SUB

SUB Initialise
DEFSTR tstr, buf, tgtFolder

    SetWindowLong(DbgForm.Handle, -8, 0)
    SetWindowLong(Application.Handle, -8, DbgForm.Handle)
    DbgForm.Show
    RichEd.AddStrings "Initialising..."

    CheckEnvironment                                                ' Get paths to rc.exe etc
    buf=SPACE$(16)
    IF fbParserVersion(VARPTR(buf)) THEN
        VERSION=FIELD$(buf,CHR$(0),1)                               ' Returns nul terminated string
        DbgForm.Caption="RapidDBG "+VERSION
    ELSE
        DbgError(9)                                                 ' Problem with the dll?
    END IF

    SELECT CASE CommandCount
        CASE 0: GetSourceFile                                       ' Gets filename, sets path etc etc
        CASE 1
            IF UCASE$(TRIM$(COMMAND$(1)))="-E" THEN
                CommandParm=EDIT_COLORS: EXIT SUB
            ELSE
                DbgError(3)                                         ' Usage help
            END IF
        CASE IS > 4                                                 ' 4th may be breakpoints
            DbgError(3)                                             ' Usage help
        CASE ELSE
            tstr=UCASE$(TRIM$(COMMAND$(1)))
            SELECT CASE tstr
                CASE "-T": CommandParm=TEST_ONLY
                CASE "-C": CommandParm=COMP_ONLY
                CASE "-R": CommandParm=RUN_USER
                CASE "-D"
                    CommandParm=RUN_DEBUG
                    PreSetBrkPoints=CheckForBreakpoints
                CASE "-X": CommandParm=CLEANUP_DBG: EXIT SUB        ' An "internal" cmd line parameter
                CASE ELSE
                    DbgError(3)                                     ' Usage help
            END SELECT

            srcFile=TRIM$(COMMAND$(2))
            srcFile=REPLACESUBSTR$(srcFile,"/","\")
            IF FileExists(srcFile)=0 THEN DbgError(4)
            
            tgtFolder=LEFT$(srcFile,RINSTR(srcFile,"\"))            ' The target directory
            tstr=srcFile-tgtFolder                                  ' User src file name+ext
            dbgFile=tgtFolder+"~"+tstr                              ' Debugger source+debug filename
            tgtFolder=LEFT$(tgtFolder,LEN(tgtFolder)-1)             ' Remove trailing '\'
            CHDIR tgtFolder                                         ' Move to the target directory
    END SELECT
END SUB

FUNCTION NewWndProc(hWnd AS long, uMsg AS long, wParam AS long, lParam AS long) AS long
DEFINT ScrollDown
DEFLNG hnd

	Result=0
    IF uMsg=DBG_MOUSEWHEEL THEN
        hnd=WindowFromPoint(Screen.MouseX, Screen.MouseY)
        IF hnd=StrGrid.Handle THEN
            ScrollDown = IIF((wParam AND &H800000)=0,False,True)
            IF ScrollDown THEN
                IF StrGrid.TopRow < StrGrid.RowCount-16 THEN            ' IF able to move down, move
                    StrGrid.TopRow=StrGrid.TopRow+1
                END IF
            ELSE
                IF StrGrid.TopRow > 1 THEN                              ' IF able to move up, move
                    StrGrid.TopRow=StrGrid.TopRow-1
                END IF
            END IF
        ELSE
            Result=CallWindowProc(OldWndProc,hwnd,uMsg,wParam,lParam)
        END IF
    ELSE
        Result=CallWindowProc(OldWndProc,hwnd,uMsg,wParam,lParam)
    END IF
END FUNCTION

SUB InitEdit
DEFINT i, j
DEFLNG bndVar
DIM TmpMem AS QMEMORYSTREAM

    IF FileExists(Application.Path+"\RapidDBG.ini")=0 THEN DbgError(2)

    BIND bndVar TO NewWndProc
    OldWndProc=SetWindowLong(EditForm.Handle, -4, SetNewCallBack_4(bndVar))

    FOR i=1 TO 10
        Label(i).Parent=EditForm
        Label(i).Left=10: Label(i).Top=i*30+6
        Label(i).Width=110: Label(i).Height=25
        Label(i).Alignment=1

        FGPanel(i).Parent=EditForm
        FGPanel(i).Left=130: FGPanel(i).Top=i*30
        FGPanel(i).Width=25: FGPanel(i).Height=25
        FGPanel(i).Caption="FG": FGPanel(i).Tag=i
        FGPanel(i).Font.Color=&Hffffff
        FGPanel(i).ShowHint=True: FGPanel(i).Hint="Set foreground"
        FGPanel(i).OnClick=EditColour

        BGPanel(i).Parent=EditForm
        BGPanel(i).Left=160: BGPanel(i).Top=i*30
        BGPanel(i).Width=25: BGPanel(i).Height=25
        BGPanel(i).Caption="BG": BGPanel(i).Tag=i+10
        BGPanel(i).ShowHint=True: BGPanel(i).Hint="Set background"
        BGPanel(i).OnClick=EditColour
    NEXT i
    Label(1).Caption="Default": Label(2).Caption="Current Line"
    Label(3).Caption="Breakpoint": Label(4).Caption="Comments"
    Label(5).Caption="RQ DEFs": Label(6).Caption="RQ Objects"
    Label(7).Caption="Types && Structs": Label(8).Caption="Variables"
    Label(9).Caption="Start-End Subs/Fns": Label(10).Caption="Sub Function Code"

    TmpMem.Close                                                    ' Reset the stream
    TmpMem.ExtractRes(Resource(3))                                  ' The demo source code
    TmpMem.Position=0
    DisplayList.LoadFromStream(TmpMem)                              ' Load the stringlist
    TmpMem.Close
    StrGrid.RowCount=DisplayList.ItemCount+1
    j=1
    FOR i=0 TO DisplayList.ItemCount-1                              ' Display the demo code
        StrGrid.Cell(0,j)=""
        StrGrid.Cell(1,j)=STR$(j)
        StrGrid.Cell(2,j)=DisplayList.Item(i)
        INC j
    NEXT i
    StrGrid.Cell(0,15)="->"                                         ' Dummy current line
    StrGrid.Cell(0,23)="B"                                          ' Dummy Breakpoint
    StrGrid.TopRow=8
    LoadColourScheme
    IdentifyEachLine
END SUB

' If default foreground is changed, update all with same colour
SUB UpdateForeGrounds(oldColor AS long, newColor AS long)
DEFINT i

    FOR i=1 TO 10
        IF FGColor(i)=oldColor THEN
            FGPanel(i).Color=newColor                           ' Update BG panel color
            FGPanel(i).Font.Color = &hffffff - newColor         ' Update text color
            FGColor(i)=newColor                                 ' Update the global array
        END IF
    NEXT i
    StrGrid.Font.Color=newColor
END SUB

' If default background is changed, update all with same colour
SUB UpdateBackGrounds(oldColor AS long, newColor AS long)
DEFINT i

    FOR i=1 TO 10
        IF BGColor(i)=oldColor THEN
            BGPanel(i).Color=newColor                           ' Update BG panel color
            BGPanel(i).Font.Color = &hffffff - newColor         ' Update text color
            BGColor(i)=newColor                                 ' Update the global array
        END IF
    NEXT i
    StrGrid.Color=newColor
END SUB

' Sender.Tag: 1-10 FGPanel: 11-20 BGPanel
SUB EditColour(Sender)
DEFINT idx, sel=Sender.Tag
DEFINT oldForGrnd=FGPanel(1).Color
DEFINT oldBKGrnd=BGPanel(1).Color
DEFSTR tstr

    idx=IIF(sel>10,sel-10,sel)
WITH ColorDlg
    .Style=cdFullOpen
    .Top=EditForm.Top+35
    .Left=EditForm.Left+FGPanel(1).Left
    tstr=REPLACESUBSTR$(Label(idx).Caption,"&&","&")                ' Code type editing
    IF sel>10 THEN
        .Caption="Background - "+tstr
        .Color=BGPanel(idx).Color                                   ' Default color for color chooser
    ELSE
        .Caption="Foreground - "+tstr
        .Color=FGPanel(idx).Color                                   ' Default color for color chooser
    END IF
    Label(idx).Font.Color=&hdf                                      ' Indicate a change
    IF .Execute THEN
        SELECT CASE sel
            CASE 1
                UpdateForeGrounds(oldForGrnd, .color)               ' Change ALL default foregrounds
            CASE 2 TO 10
                FGPanel(idx).Color=.Color                           ' Update FG panel color
                FGPanel(idx).Font.Color =  &hffffff - .color        ' Update text color
                FGColor(idx)=.color                                 ' Update the global array
            CASE 11
                UpdateBackGrounds(oldBKGrnd, .color)                ' Change ALL default backgrounds
            CASE 12 TO 20
                BGPanel(idx).Color=.Color                           ' Update BG panel color
                BGPanel(idx).Font.Color = &hffffff - .color         ' Update text color
                BGColor(idx)=.color                                 ' Update the global array
        END SELECT
        StrGrid.Repaint
    END IF
    Label(idx).Font.Color=0                                         ' Back to default
END WITH
END SUB

SUB ColourThisLine(Col, Row, State, Rect)
DEFSTR str
DEFLNG fColor, bColor

WITH StrGrid
    IF Row>0 THEN
        IF .Cell(0,Row)="B" THEN
            fColor=FGColor(3): bColor=BGColor(3)                    ' Brkpoint
        ELSE
            str=DisplayList.Item(Row)
            SELECT CASE str
                CASE "C": fColor=FGColor(4): bColor=BGColor(4)      ' Comments
                CASE "R": fColor=FGColor(5): bColor=BGColor(5)      ' $DEFs & CONSTs
                CASE "E": fColor=FGColor(6): bColor=BGColor(6)      ' Create..End Create
                CASE "T": fColor=FGColor(7): bColor=BGColor(7)      ' Type..End Type & structs
                CASE "V": fColor=FGColor(8): bColor=BGColor(8)      ' Variables
                CASE "S": fColor=FGColor(9): bColor=BGColor(9)      ' Start & end of subs/fns
                CASE "F": fColor=FGColor(10): bColor=BGColor(10)    ' Code of subs/fns
                CASE ELSE
                    fColor=FGColor(1): bColor=BGColor(1)            ' Remainder
            END SELECT
        END IF
        str=.Cell(Col,Row)+SPACE$(250)                              ' With full line padding
        .TextOut(Rect.Left+2,Rect.Top+2,str,fColor,bColor)

        IF .Cell(0,Row)="->" THEN                                   ' Is this the current line?
            str=RTRIM$(FIELD$(str,"'",1))+" "                       ' Remove any comment, add small padding
            .TextOut(Rect.Left+2,Rect.Top+2,str,FGColor(2),BGColor(2))      ' OverWrite code ONLY
        END IF
    END IF
END WITH
END SUB

SUB LoadColourScheme
DEFSTR tstr, token
DEFINT i
DEFLNG col

    FOR i=1 TO 10: FGColor(i)=0: BGColor(i)=&hffffff: NEXT i        ' Set initial defaults    
    
    tstr=Application.Path+"\RapidDBG.ini"
    DisplayList.Clear
    DisplayList.LoadFromFile(tstr)
    IF DisplayList.ItemCount<4 THEN DBGError(2)

    FOR i=4 TO DisplayList.ItemCount-1
        tstr=RTRIM$(FIELD$(DisplayList.Item(i),"'",1))              ' Remove any comments
        tstr=REPLACESUBSTR$(tstr,CHR$(9)," ")
        token=UCASE$(LTRIM$(FIELD$(tstr,"=",1)))
        col=VAL(CONVBASE$(FIELD$(tstr,"=",2),16,10))
        SELECT CASE token
            CASE "FG_DEFAULT": FGColor(1)=col: FGPanel(1).Color=col
            CASE "BG_DEFAULT": BGColor(1)=col: BGPanel(1).Color=col
            CASE "FG_CURLINE": FGColor(2)=col: FGPanel(2).Color=col
            CASE "BG_CURLINE": BGColor(2)=col: BGPanel(2).Color=col
            CASE "FG_BRKPOINT": FGColor(3)=col: FGPanel(3).Color=col
            CASE "BG_BRKPOINT": BGColor(3)=col: BGPanel(3).Color=col
            CASE "FG_COMMENT": FGColor(4)=col: FGPanel(4).Color=col
            CASE "BG_COMMENT": BGColor(4)=col: BGPanel(4).Color=col
            CASE "FG_RQDEFS": FGColor(5)=col: FGPanel(5).Color=col
            CASE "BG_RQDEFS": BGColor(5)=col: BGPanel(5).Color=col
            CASE "FG_RQCREATE": FGColor(6)=col: FGPanel(6).Color=col
            CASE "BG_RQCREATE": BGColor(6)=col: BGPanel(6).Color=col
            CASE "FG_RQTYPES": FGColor(7)=col: FGPanel(7).Color=col
            CASE "BG_RQTYPES": BGColor(7)=col: BGPanel(7).Color=col
            CASE "FG_RQVARS": FGColor(8)=col: FGPanel(8).Color=col
            CASE "BG_RQVARS": BGColor(8)=col: BGPanel(8).Color=col
            CASE "FG_SUBFNS": FGColor(9)=col: FGPanel(9).Color=col
            CASE "BG_SUBFNS": BGColor(9)=col: BGPanel(9).Color=col
            CASE "FG_SUBFNCODE": FGColor(10)=col: FGPanel(10).Color=col
            CASE "BG_SUBFNCODE": BGColor(10)=col: BGPanel(10).Color=col
        END SELECT
    NEXT i
    StrGrid.Color=BGColor(1)
    DisplayList.Clear
END SUB

FUNCTION IsLineVariable(srcLine AS string) AS integer
    Result=FALSE
    IF LEFT$(srcLine,8)="DEFBYTE " OR LEFT$(srcLine,9)="DEFSHORT " _
            OR LEFT$(srcLine,7)="DEFINT " OR LEFT$(srcLine,7)="DEFLNG "  _
            OR LEFT$(srcLine,7)="DEFSNG " OR LEFT$(srcLine,7)="DEFDBL " _
            OR LEFT$(srcLine,8)="DEFWORD " OR LEFT$(srcLine,7)="DEFSTR " _
            OR LEFT$(srcLine,4)="DIM " OR LEFT$(srcLine,7)="STATIC " THEN
        Result=TRUE
    END IF
END FUNCTION

' "C"=comments: "R"=RQ specific: "S"=Start & end Sub/Fn: "F"=Code in Sub/fn
' "T"=Type..End Type: "E"=Create..End Create: "V"=Variables: ""=Default
SUB IdentifyEachLine
DEFINT i, inSubFn, inType, inCreate, lev
DEFSTR tstr, res

    inSubFn=0: inType=0: inCreate=0: lev=0
    DisplayList.Clear
    DisplayList.AddItems ""                                         ' Don't use index 0
    FOR i=1 TO StrGrid.RowCount-1
        tstr=UCASE$(LTRIM$(StrGrid.Cell(2,i)))                      ' Each src line
        IF LEFT$(tstr,7)="PUBLIC " THEN tstr=LTRIM$(tstr-"PUBLIC ")
        IF (LEFT$(tstr,1)="'") OR (LEFT$(tstr,4)="REM ") THEN
            res="C"                                                 ' "C"omments
        ELSEIF (LEFT$(tstr,1)="$") OR (LEFT$(tstr,6)="CONST ") OR _
                                            (LEFT$(tstr,8)="DECLARE ") THEN
            res="R"                                                 ' "R"Q specific
        ELSE
            tstr=RTRIM$(FIELD$(tstr,"'",1))                                     ' Remove any comments
            IF FIELD$(tstr," ",1)="PUBLIC" THEN tstr=LTRIM$(tstr-"PUBLIC ")     ' Remove this
            ' Each remaining line
            IF (LEFT$(tstr,4)="SUB ") OR (LEFT$(tstr,5)="SUBI ") OR _
                    (LEFT$(tstr,9)="FUNCTION ") OR (LEFT$(tstr,10)="FUNCTIONI ") THEN
                res="S": inSubFn=1
            ELSEIF (LEFT$(tstr,7)="END SUB") OR (LEFT$(tstr,12)="END FUNCTION") THEN
                res="S": inSubFn=0
            ELSEIF IsLineVariable(tstr) THEN                        ' Variables
                res="V"
            ELSEIF LEFT$(tstr,7)="CREATE " THEN
                res="E": inCreate=1: INC lev
            ELSEIF LEFT$(tstr,10)="END CREATE" THEN
                res="E": DEC lev: IF lev=0 THEN inCreate=0
            ELSEIF (LEFT$(tstr,5)="TYPE ") OR (LEFT$(tstr,7)="STRUCT ") THEN
                res="T": inType=1
            ELSEIF (LEFT$(tstr,8)="END TYPE") OR (LEFT$(tstr,10)="END STRUCT") THEN
                res="T": inType=0
            ELSEIF inSubFn THEN
                res="F"
            ELSEIF inType THEN
                res="T"
            ELSEIF inCreate THEN
                res="E"
            ELSE
                res=""
            END IF
        END IF
        DisplayList.AddItems res
    NEXT i
END SUB

SUB UpdateIniFile
DEFINT i,j
DEFSTR tstr
DIM tmpList AS QStringList

    tmpList.Clear
    tmpList.LoadFromFile(Application.Path+"\RapidDBG.ini")
    DisplayList.Clear
    FOR i=0 TO 4
        DisplayList.AddItems tmpList.Item(i)
    NEXT i
    DisplayList.AddItems "FG_DEFAULT=&H"+RIGHT$(HEX$(FGColor(1)),6)
    DisplayList.AddItems "BG_DEFAULT=&H"+RIGHT$(HEX$(BGColor(1)),6)
    DisplayList.AddItems "FG_CURLINE=&H"+RIGHT$(HEX$(FGColor(2)),6)
    DisplayList.AddItems "BG_CURLINE=&H"+RIGHT$(HEX$(BGColor(2)),6)
    DisplayList.AddItems "FG_BRKPOINT=&H"+RIGHT$(HEX$(FGColor(3)),6)
    DisplayList.AddItems "BG_BRKPOINT=&H"+RIGHT$(HEX$(BGColor(3)),6)
    DisplayList.AddItems "FG_COMMENT=&H"+RIGHT$(HEX$(FGColor(4)),6)
    DisplayList.AddItems "BG_COMMENT=&H"+RIGHT$(HEX$(BGColor(4)),6)
    DisplayList.AddItems "FG_RQDEFS=&H"+RIGHT$(HEX$(FGColor(5)),6)
    DisplayList.AddItems "BG_RQDEFS=&H"+RIGHT$(HEX$(BGColor(5)),6)
    DisplayList.AddItems "FG_RQCREATE=&H"+RIGHT$(HEX$(FGColor(6)),6)
    DisplayList.AddItems "BG_RQCREATE=&H"+RIGHT$(HEX$(BGColor(6)),6)
    DisplayList.AddItems "FG_RQTYPES=&H"+RIGHT$(HEX$(FGColor(7)),6)
    DisplayList.AddItems "BG_RQTYPES=&H"+RIGHT$(HEX$(BGColor(7)),6)
    DisplayList.AddItems "FG_RQVARS=&H"+RIGHT$(HEX$(FGColor(8)),6)
    DisplayList.AddItems "BG_RQVARS=&H"+RIGHT$(HEX$(BGColor(8)),6)
    DisplayList.AddItems "FG_SUBFNS=&H"+RIGHT$(HEX$(FGColor(9)),6)
    DisplayList.AddItems "BG_SUBFNS=&H"+RIGHT$(HEX$(BGColor(9)),6)
    DisplayList.AddItems "FG_SUBFNCODE=&H"+RIGHT$(HEX$(FGColor(10)),6)
    DisplayList.AddItems "BG_SUBFNCODE=&H"+RIGHT$(HEX$(BGColor(10)),6)

    DisplayList.SaveToFile(Application.Path+"\RapidDBG.ini")
    DoEvents
END SUB

SUB ShowHelp
DEFSTR tstr

    tstr="Debugger Color Editor"+CRLF
    tstr=tstr+"Part of RapidDBG Source-line Debugger for RapidQ "+CRLF+CRLF
    tstr=tstr+"Colorize the source code screen to suit your preferences."+CRLF
    tstr=tstr+"For best results, use a dark foreground with an"+CRLF
    tstr=tstr+"extra light background color (or an opposite combination)."+CRLF+CRLF
    tstr=tstr+"The background color will extend across the entire line"+CRLF
    tstr=tstr+"of code (with the exception of "+QU+"Current Line"+QU+" setting)."+CRLF+CRLF
    tstr=tstr+"To restore defaults, delete the RapidDBG.ini file and"+CRLF
    tstr=tstr+"then run RapidDBG.exe again"+CRLF
    MessageBox (tstr, "Help",&h40)
END SUB

SUB EditAbort
    Application.Terminate
END SUB

SUB EditExit
    UpdateIniFile
    Application.Terminate
END SUB
